import os
import datetime
import numbers
from itertools import zip_longest
from typing import Sequence, Optional, Dict, Union


def mca_data_to_spec_string(
    mca: Sequence[float],
    title: Optional[str] = None,
    filename: Optional[str] = None,
    date: Optional[Union[str, datetime.datetime, datetime.date]] = None,
    calibration: Optional[Sequence[float]] = None,
    detector_name: Optional[str] = None,
    metadata: Optional[Dict] = None,
) -> str:
    """Official file format specs: https://certif.com/downloads/css_docs/spec_man.pdf

    :param mca: MCA spectrum
    :param title: scan title
    :param filename: name by which the file will be created
    :param date: start date of the scan
    :param calibration: MCA calibration as the sequence [zero, gain, quad],
                        all of them optional so that
                        energy = zero + gain*channels + quad*channels^2
    :param detector_name: name of the MCA detector
    :param metadata: saved as comments
    :returns: SPEC serialized string
    """
    nchan_per_line = 16
    nchan_tot = len(mca)

    if title is None:
        title = "ct"
    if filename is None:
        filename = "unspecified"
    if detector_name is None:
        detector_name = "MCA0"
    if date is None:
        date = datetime.datetime.now()
    calib = [0, 1, 0]  # zero, gain, quad
    if calibration:
        if len(calibration) > 3:
            raise ValueError("MCA calibration requires 3 only coefficients")
        calib = [
            p if p is not None else pdefault
            for p, pdefault in zip_longest(calibration, calib)
        ]

    header = [f"#F {filename}", f"#D {date}", "", f"#S {title}", f"#D {date}"]
    if metadata:
        for k, v in metadata.items():
            header.append(f"#C {k} = {v}")
    header.append("#N 1")
    header.append(f"#@MCA {nchan_per_line}C")
    header.append(f"#@CHANN {nchan_tot} 0 {nchan_tot-1} 1")
    header.append(f"#@CALIB {' '.join(map(str, calib))}")
    header.append("#@MCA_NB 1")
    header.append(f"#L {detector_name}")

    mcastring = "\n".join(header)

    mcastring += "\n@A"
    if isinstance(mca[0], numbers.Integral):
        fmt = " %d"
    else:
        # fmt = " %.4f"
        fmt = " %.8g"
    for idx in range(0, nchan_tot, nchan_per_line):
        if idx + nchan_per_line - 1 < nchan_tot:
            for i in range(0, nchan_per_line):
                mcastring += fmt % mca[idx + i]
            if idx + nchan_per_line != nchan_tot:
                mcastring += "\\"
        else:
            for i in range(idx, nchan_tot):
                mcastring += fmt % mca[i]
        mcastring += "\n"
    return mcastring


def save_as_spec(filename, mca: Sequence[float], **kwargs) -> None:
    dirname = os.path.dirname(filename)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    content = mca_data_to_spec_string(mca, filename=filename, **kwargs)
    with open(filename, "w") as f:
        f.write(content)



import numpy

from ..tasks import io


def test_io_spec():
    mca = numpy.arange(6)
    mca_string = io.mca_data_to_spec_string(mca, date="start_date")
    expected = [
        "#F unspecified",
        "#D start_date",
        "",
        "#S ct",
        "#D start_date",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 6 0 5 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(16)
    mca_string = io.mca_data_to_spec_string(mca, date="start_date")
    expected = [
        "#F unspecified",
        "#D start_date",
        "",
        "#S ct",
        "#D start_date",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 16 0 15 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(32)
    mca_string = io.mca_data_to_spec_string(mca, date="start_date")
    expected = [
        "#F unspecified",
        "#D start_date",
        "",
        "#S ct",
        "#D start_date",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 32 0 31 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(33)
    mca_string = io.mca_data_to_spec_string(mca, date="start_date")
    expected = [
        "#F unspecified",
        "#D start_date",
        "",
        "#S ct",
        "#D start_date",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 33 0 32 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\\",
        " 32",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(33)
    mca_string = io.mca_data_to_spec_string(
        mca, date="start_date", metadata={"a": 1, "b": 2}
    )
    expected = [
        "#F unspecified",
        "#D start_date",
        "",
        "#S ct",
        "#D start_date",
        "#C a = 1",
        "#C b = 2",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 33 0 32 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\\",
        " 32",
        "",
    ]
    assert mca_string.split("\n") == expected



import h5py
import numpy
import pytest

from ...tasks.hdf5_to_spec import Hdf5ToSpec


def test_hdf5_to_spec(tmp_path):
    filename = tmp_path / "RAW_DATA" / "bliss_dataset.h5"
    filename.parent.mkdir()
    filename = str(filename)
    output_filename = str(tmp_path / "PROCESSED_DATA" / "bliss_dataset.mca")

    nscans = 3
    nchannels = 10
    with h5py.File(filename, "w") as nxroot:
        for scan in range(1, nscans + 1):
            nxroot[f"/{scan}.1/start_time"] = "start_time"
            nxroot[f"/{scan}.1/title"] = "timescan 0.1"
            nxroot[f"/{scan}.1/measurement/Counts"] = numpy.full((1, nchannels), scan)
            nxroot[f"/{scan}.1/end_time"] = "end_time"

    inputs = {
        "filename": filename,
        "scan_numbers": list(range(1, nscans + 1)),
        "output_filename": output_filename,
        "mca_calibration": [0.1, 0.2],
    }
    task = Hdf5ToSpec(inputs=inputs)
    task.run()

    output_filenames = [
        str(tmp_path / "PROCESSED_DATA" / f"bliss_dataset_{i:02d}.mca")
        for i in range(1, nscans + 1)
    ]
    assert task.get_output_values() == {"output_filenames": output_filenames}

    for i, filename in enumerate(output_filenames, 1):
        with open(filename, "r") as f:
            lines = [s.rstrip() for s in f.readlines()]

            expected = [
                f"#F {filename}",
                "#D start_time",
                "",
                f"#S {i} timescan 0.1",
                "#D start_time",
                "#C Finished = end_time",
                "#N 1",
                "#@MCA 16C",
                f"#@CHANN {nchannels} 0 {nchannels-1} 1",
                "#@CALIB 0.1 0.2 0",
                "#@MCA_NB 1",
                "#L Counts",
                "@A " + " ".join(map(str, [i] * 10)),
            ]
            assert lines == expected


def test_hdf5_to_spec_failed(tmp_path):
    filename = tmp_path / "RAW_DATA" / "bliss_dataset.h5"
    filename.parent.mkdir()
    filename = str(filename)
    output_filename = str(tmp_path / "PROCESSED_DATA" / "bliss_dataset.mca")

    nscans = 1
    nchannels = 10
    with h5py.File(filename, "w") as nxroot:
        for scan in range(1, nscans + 1):
            nxroot[f"/{scan}.1/start_time"] = "start_time"
            nxroot[f"/{scan}.1/title"] = "timescan 0.1"
            nxroot[f"/{scan}.1/measurement/Counts"] = numpy.full((1, nchannels), scan)
            nxroot[f"/{scan}.1/end_time"] = "end_time"

    inputs = {
        "filename": filename,
        "scan_numbers": list(range(1, nscans + 3)),
        "output_filename": output_filename,
        "retry_timeout": 0.1,
    }
    task = Hdf5ToSpec(inputs=inputs)
    with pytest.raises(
        RuntimeError, match=r"^Failed scans \(see logs why\): \[2, 3\]$"
    ):
        task.run()



import os
import logging
from typing import Optional, Sequence, Any, Union

from ewokscore.task import Task
from blissdata.h5api import dynamic_hdf5

from .io import save_as_spec

logger = logging.getLogger(__name__)


class Hdf5ToSpec(
    Task,
    input_names=["filename", "output_filename"],
    optional_input_names=[
        "scan_numbers",
        "retry_timeout",
        "retry_period",
        "mca_counter",
        "mca_calibration",
    ],
    output_names=["output_filenames"],
):
    def run(self):
        filename: str = self.inputs.filename
        scan_numbers: Optional[Sequence[int]] = self.get_input_value(
            "scan_numbers", None
        )
        mca_counter: str = self.get_input_value("mca_counter", "Counts")
        retry_timeout: Optional[float] = self.get_input_value("retry_timeout", None)
        retry_period: Optional[float] = self.get_input_value("retry_period", 1)
        mca_calibration: Optional[float] = self.get_input_value("mca_calibration", None)

        output_filename: str = self.inputs.output_filename
        dirname = os.path.dirname(output_filename)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        output_filename, ext = os.path.splitext(output_filename)

        output_filenames = list()
        failed_scans = list()
        with dynamic_hdf5.File(
            filename, retry_timeout=retry_timeout, retry_period=retry_period
        ) as nxroot:
            if scan_numbers:
                scan_names = [f"{scannr}.1" for scannr in scan_numbers]
            else:
                scan_names = nxroot

            for scan_name in scan_names:
                scan_number = int(float(scan_name))

                try:
                    title = _asstr(nxroot[f"/{scan_name}/title"][()])
                    start_time = _asstr(nxroot[f"/{scan_name}/start_time"][()])
                    end_time = _asstr(nxroot[f"/{scan_name}/end_time"][()])
                    mca = nxroot[f"/{scan_name}/measurement/{mca_counter}"][-1]
                except Exception as e:
                    failed_scans.append(scan_number)
                    logger.error(
                        "Processing scan %s::/%s failed (%s)", filename, scan_name, e
                    )
                    continue

                scan_output_filename = f"{output_filename}_{scan_number:02d}{ext}"
                metadata = {"Finished": end_time}

                save_as_spec(
                    scan_output_filename,
                    mca,
                    title=f"{scan_number} {title}",
                    date=start_time,
                    detector_name=mca_counter,
                    metadata=metadata,
                    calibration=mca_calibration,
                )
                output_filenames.append(scan_output_filename)

        if failed_scans:
            raise RuntimeError(f"Failed scans (see logs why): {failed_scans}")

        self.outputs.output_filenames = output_filenames


def _asstr(data: Union[str, bytes]) -> Any:
    if isinstance(data, bytes):
        return data.decode()
    return data
